/*
 * This file is a part of morm
 * and is distributed under the 2-Clause BSD licence.
 * Author: Tomasz Sowa <t.sowa@ttmath.org>
 */

/*
 * Copyright (c) 2018, Tomasz Sowa
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 * 1. Redistributions of source code must retain the above copyright notice,
 * this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 * notice, this list of conditions and the following disclaimer in the
 * documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 * LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 *
 */

#ifndef headerfile_morm_modelconnector
#define headerfile_morm_modelconnector

#include "baseexpression.h"
#include "dbconnector.h"
#include "flatexpression.h"
#include "flatconnector.h"


namespace morm
{
class Model;


template<typename ModelClass>
class Finder;


class ModelConnector
{

public:

	ModelConnector();
	virtual ~ModelConnector();

	// FIX ME
	// add c-copy ctr (allocate a new stream and expression)

	virtual void set_stream(PT::TextStream & stream);
	virtual PT::TextStream * get_stream();

	virtual void set_flat_connector(FlatConnector & flat_connector);

	virtual void set_db_connector(DbConnector & db_connector);
	virtual DbConnector * get_db_connector();


	virtual void set_db_expression_output_type(int output_type);

	virtual void to_text(PT::TextStream & stream, Model & model);


	virtual void generate_select_columns(PT::TextStream & stream, Model & model);
	virtual void generate_insert_query(PT::TextStream & stream, Model & model);
	virtual void generate_update_query(PT::TextStream & stream, Model & model);

	virtual void insert(Model & model);
	virtual void update(Model & model);

	virtual void map_values_from_query(Model & model);

	virtual void clear_values(Model & model);


//	template<typename ModelClass>
//	Finder<ModelClass> & find()
//	{
//		allocate_default_stream_if_needed();
//		return Finder<ModelClass>(this);
//	}


protected:

	int model_connector_mode;

	FlatConnector * flat_connector;
	DbConnector * db_connector;

	PT::TextStream * out_stream; // IMPROVE ME give here an interface to the base stream (implement him)
	bool out_stream_allocated;

	//BaseExpression * expression_callback;
	//DbConnector * db_connector_callback;

	void allocate_default_stream();
	void allocate_default_stream_if_needed();
	void deallocate_stream();


	template<typename FieldValue>
	void field(const wchar_t * field_name, FieldValue & field_value, bool insertable, bool updatable, bool is_primary_key)
	{
		// may we need only db_connector_callback?

//		if( expression_callback )
//		{
//			expression_callback->field(field_name, field_value, insertable, updatable, is_primary_key);
//		}

//		if( db_connector_callback )
//		{
//			db_connector_callback->get_value_by_field_name(field_name, field_value);
//		}

		if( model_connector_mode == MORM_MODEL_CONNECTOR_MODE_GENERATING_FLAT_EXPRESSION && flat_connector )
		{
			FlatExpression * flat_expression = flat_connector->get_expression();

			if( flat_expression )
			{
				flat_expression->field(field_name, field_value, insertable, updatable, is_primary_key);
			}
		}

		if( model_connector_mode == MORM_MODEL_CONNECTOR_MODE_GENERATING_DB_EXPRESSION && db_connector )
		{
			DbExpression * db_expression = db_connector->get_expression();

			if( db_expression )
			{
				db_expression->field(field_name, field_value, insertable, updatable, is_primary_key);
			}
		}

		if( model_connector_mode == MORM_MODEL_CONNECTOR_MODE_READING_VALUE_FROM_DB && db_connector )
		{
			db_connector->get_value_by_field_name(field_name, field_value);
		}

		if( model_connector_mode == MORM_MODEL_CONNECTOR_MODE_CLEARING_VALUE && db_connector )
		{
			// IMPROVE ME those clearing should be moved to a better place
			db_connector->clear_value(field_value);
		}
	}

	template<typename ModelClass>
	void field_list(const wchar_t * field_name, std::list<ModelClass> & field_list, bool insertable = true, bool updatable = true, bool is_primary_key = false)
	{
		if( model_connector_mode == MORM_MODEL_CONNECTOR_MODE_GENERATING_FLAT_EXPRESSION && flat_connector && out_stream )
		{
			FlatExpression * flat_expression = flat_connector->get_expression();

			if( flat_expression )
			{
				flat_expression->field_list(field_name, field_list, insertable, updatable, is_primary_key);
			}
		}
	}


	template<typename FieldValue>
	void get_last_sequence(const wchar_t * sequence_table_name, FieldValue & field_value)
	{
		if( db_connector )
		{
			db_connector->get_last_sequence(sequence_table_name, field_value);
		}
	}


	/*
	 * Model is using field() method
	 */
	friend class Model;
	template<typename ModelClass> friend class Finder;

};

}

#endif
